#include <iostream>

#include "hdnum.hh"

#include "single_shooting.hh"


/*Example class for a differential equation of the form

  u''(t) = f(t,u(t),u'(t)), a<=t<=b
  u(a) = u_a
  u(b) = u_b
*/
template <typename T, typename N=T>
class SecondOrderODE
{
public:
  /** \brief export size_type */
  typedef std::size_t size_type;

  /** \brief export time_type */
  typedef T time_type;

  /** \brief export number_type */
  typedef N number_type;

  // Function f from the right hand side of second order ODE. Assume
  // x[0]=u and x[1]=u'.
  void f(T t, const hdnum::Vector<N>& x, N& result) const
  {
    // TODO
  }

  // Derivatives of f
  //
  // result[0] = \partial_u f
  // result[1] = \partial_u' f
  void f_x(T t, const hdnum::Vector<N>& x, hdnum::Vector<N>& result) const
  {
    // TODO
  }

  // result[0] = u_a
  // result[1] = u_b
  void boundary_values (hdnum::Vector<N>& result) const
  {
    // TODO
  }
};

// Initial value problem needed for evaluating nonlinear function in Newton
template<typename T, typename N=T>
class Model
{
public:
  /** \brief export size_type */
  typedef std::size_t size_type;

  /** \brief export time_type */
  typedef T time_type;

  /** \brief export number_type */
  typedef N number_type;

  //! constructor stores parameter
  Model (const SecondOrderODE<T>& second_order_ode_)
    : second_order_ode(second_order_ode_), s(0.0)
  {}

  //! return number of componentes for the model
  std::size_t size () const
  {
    return 2;
  }

  void set_s(N s_)
  {
    s = s_;
  }

  //! set initial state including time value
  void initialize (T& t0, hdnum::Vector<N>& x0) const
  {
    t0 = 0;
    hdnum::Vector<N> boundary(2);
    second_order_ode.boundary_values(boundary);
    x0[0] = boundary[0];
    x0[1] = s;
  }

  //! model evaluation
  void f (const T& t, const hdnum::Vector<N>& x, hdnum::Vector<N>& result) const
  {
    result[0] = x[1];
    N f_eval;
    second_order_ode.f(t, x, f_eval);
    result[1] = f_eval;
  }

  //! Jacobi matrix
  void f_x (T t, const hdnum::Vector<N>& , hdnum::DenseMatrix<N>& result) const
  {
    // Not implemented
    assert(false);
  }

private:
  SecondOrderODE<T> second_order_ode;
  N s;
};


// Initial value problem needed for evaluation of derivative of
// nonlinear function in Newton
template<typename T, typename N=T>
class DerivativeModel
{
public:
  /** \brief export size_type */
  typedef std::size_t size_type;

  /** \brief export time_type */
  typedef T time_type;

  /** \brief export number_type */
  typedef N number_type;

  //! constructor stores parameter
  DerivativeModel (const SecondOrderODE<T>& second_order_ode_)
    : second_order_ode(second_order_ode_), y(2)
  {}

  //! return number of componentes for the model
  std::size_t size () const
  {
    return 2;
  }

  void set_y (const hdnum::Vector<N>& y_)
  {
    y[0] = y_[0];
    y[1] = y_[1];
  }

  //! set initial state including time value
  void initialize (T& t0, hdnum::Vector<N>& x0) const
  {
    t0 = 0;
    x0[0] = 0;
    x0[1] = 1;
  }

  //! model evaluation
  void f (T t, const hdnum::Vector<N>& x, hdnum::Vector<N>& result) const
  {
    result[0] = x[1];
    hdnum::Vector<N> f_deriv(2);
    second_order_ode.f_x(t, y, f_deriv);
    result[1] = f_deriv[0]*x[0] + f_deriv[1]*x[1];
  }

  //! Jacobi matrix
  void f_x (T t , const hdnum::Vector<N>& , hdnum::DenseMatrix<N>& result) const
  {
    // Not implemented
    assert(false);
  }

private:
  SecondOrderODE<T> second_order_ode;
  hdnum::Vector<N> y;
};



int main()
{
  // Setup second order ODE model
  typedef double Number;
  typedef SecondOrderODE<Number> SecondOrderODE;
  SecondOrderODE second_order_ode;

  // Setup initial value problem for F evaluation of Newton
  typedef Model<Number> Model;
  Model model(second_order_ode);

  // Setup initial value problem for F' evaluation of Newton
  typedef DerivativeModel<Number> DerivativeModel;
  DerivativeModel derivative_model(second_order_ode);

  // Initial value problem solver
  typedef hdnum::Heun2<Model> ModelSolver;
  ModelSolver model_solver(model);
  typedef hdnum::Heun2<DerivativeModel> DerivativeModelSolver;
  DerivativeModelSolver derivative_model_solver(derivative_model);

  // Shooting method
  hdnum::Vector<Number> boundary(2);
  second_order_ode.boundary_values(boundary);
  Number u_b = boundary[1];
  Number T(M_PI);
  Number dt(1e-4);
  typedef SingleShooting<Model, DerivativeModel> Solver;
  Solver solver(model, derivative_model, u_b, T, dt, true);
  solver.solve();

  return 0;
}
